"""Wordly.

A simple python client that requests dictionary definitions
from a server that implements Dictionary Server Protocol.
"""

from __future__ import annotations

from wordly.words import Word

__app_name__ = "Wordly"
__app_description__ = (
    "A simple python client that requests dictionary definitions"
    " from a server that implements Dictionary Server Protocol."
)
__epilog__ = "Stay in the know."
__version__ = "0.1.2"

__all__ = ["Word"]



"""DICT client."""

from __future__ import annotations

import asyncio

from wordly.parser import DictParser
from wordly.status_codes import Status


class DictClient:
    """Client."""

    def __init__(self, hostname: str = "dict.org", port: int = 2628) -> None:
        """Initialize."""
        self.hostname = hostname
        self.port = port
        self.line_reader = DictParser()
        self.parsers = [self.line_reader]
        self.reader: asyncio.StreamReader | None = None
        self.writer: asyncio.StreamWriter | None = None
        self.READ_BYTES = 1024

    def __repr__(self) -> str:
        """Return string representation of object."""
        return f"{self.__class__.__name__}({self.hostname=}, {self.port=})"

    async def __aenter__(self):
        """Enter method for async context manager."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """Exit method for async context manager."""
        await self.disconnect()

    async def connect(self):
        """Upon successful connection a status code of 220 is expected."""
        self.reader, self.writer = await asyncio.open_connection(
            self.hostname, self.port
        )
        connection_info = await self.reader.read(self.READ_BYTES)
        self.line_reader.feed(connection_info)

        if Status.INITIAL_CONNECTION.name in self.line_reader.mapping:
            return self.reader, self.writer

        raise ConnectionError(f"Could not connect to: {self.hostname=}, {self.port=}")

    async def disconnect(self):
        """Close client connection."""
        self.writer.write(b"QUIT\r\n")
        await self.writer.drain()
        while Status.CLOSING_CONNECTION.name not in self.line_reader.mapping:
            closing_data = await self.reader.read(self.READ_BYTES)
            for line_reader in self.parsers:
                line_reader.feed(closing_data)
        self.writer.close()
        await self.writer.wait_closed()

    async def _send(self, command: bytes) -> DictParser:
        """Return line reader given a command."""
        if None in (self.reader, self.writer):
            self.reader, self.writer = await self.connect()
        else:
            new_line_reader = DictParser()
            new_line_reader.mapping[Status.INITIAL_CONNECTION.name] = (
                self.line_reader.mapping[Status.INITIAL_CONNECTION.name]
            )
            self.line_reader = new_line_reader
            self.parsers.append(self.line_reader)

        self.writer.write(command)
        await self.writer.drain()

        while (
            Status.COMMAND_COMPLETE.name not in self.line_reader.mapping
            and Status.NO_MATCH.name not in self.line_reader.mapping
        ):
            command_data = await self.reader.read(self.READ_BYTES)
            self.line_reader.feed(command_data)
        return self.line_reader

    async def define(self, word: str, database: str = "!") -> DictParser:
        """Return line reader given word and database."""
        command = f"DEFINE {database} {word}\r\n".encode()
        return await self._send(command)

    async def help(self) -> DictParser:
        """Return line reader with helpful information."""
        command = b"HELP\r\n"
        return await self._send(command)

    async def match(
        self, word: str, database: str = "*", strategy: str = "."
    ) -> DictParser:
        """Match a word in a database using a strategy."""
        command = f"MATCH {database} {strategy} {word}".encode()
        return await self._send(command)

    async def show(self, option: str = "DB") -> DictParser:
        """Show more information."""
        command = f"SHOW {option}".encode()
        return await self._send(command)



"""Command line application module.

Parses command line arguments passed to the app and prints word definitions.
"""

from __future__ import annotations

import argparse
import asyncio
from collections.abc import Sequence

from wordly import __app_description__, __app_name__, __epilog__, __version__
from wordly.words import Word


async def print_definition(word: str, hostname: str, port: int) -> None:
    """Print definitions to stdout."""
    w = Word(word, hostname=hostname, port=port)
    definition = await w.adefinition
    print(definition)


async def main(argv: Sequence[str] | None = None) -> None:
    """Run command line application."""
    parser = argparse.ArgumentParser(
        prog=__app_name__,
        description=__app_description__,
        epilog=__epilog__,
    )
    opt = parser.add_argument

    opt("words", nargs="+", help="One or more words to define.")
    opt("-v", "--version", version=__version__, action="version")
    opt(
        "-p",
        "--port",
        type=int,
        default=2628,
        help="Specify the port number. Default: 2628",
    )
    opt(
        "-H",
        "--hostname",
        type=str,
        default="dict.org",
        help="Specify the server. Default: dict.org",
    )

    args = vars(parser.parse_args(argv))

    tasks = [
        print_definition(word, args["hostname"], args["port"]) for word in args["words"]
    ]

    await asyncio.gather(*tasks)



"""Test DICT parser."""

from __future__ import annotations

import unittest

from tests.constants import HELP_OUTPUT, PROGRAMMING_DEFINITION
from wordly.parser import DictParser
from wordly.status_codes import Status


class TestParser(unittest.TestCase):
    """Parser tests."""

    def setUp(self) -> None:
        """Set up fixtures for each test."""
        self.line_reader = DictParser()

    def test_help(self):
        """Should parse help byte stream."""
        self.line_reader.feed(HELP_OUTPUT)

        self.assertEqual(
            self.line_reader.mapping[Status.INITIAL_CONNECTION.name],
            bytearray(b"banner info from example.org\n"),
        )
        self.assertEqual(
            self.line_reader.mapping[Status.COMMAND_COMPLETE.name], bytearray(b"ok\n")
        )
        self.assertEqual(
            self.line_reader.mapping[Status.CLOSING_CONNECTION.name],
            bytearray(b"bye [d/m/c = 0/0/0; 0.000r 0.000u 0.000s]"),
        )

    def test_define(self):
        """Should parse define byte stream."""
        self.line_reader.feed(PROGRAMMING_DEFINITION)

        self.assertEqual(
            self.line_reader.mapping[Status.INITIAL_CONNECTION.name],
            bytearray(b"banner information contained here\n"),
        )
        self.assertEqual(
            self.line_reader.mapping[Status.CLOSING_CONNECTION.name],
            bytearray(b"bye [d/m/c = 0/0/0; 0.000r 0.000u 0.000s]"),
        )
